const std = @import("std");

pub fn ioctl(handle: std.os.fd_t, request: u32, arg: usize) !void {
    const rc = std.os.linux.ioctl(handle, request, arg);
    try throwReturnCodeError(rc);
    return;
}

pub fn throwReturnCodeError(return_code: usize) LinuxGenericError!void {
    // zig fmt: off
    switch (std.os.errno(return_code)) {
        .SUCCESS               => return,
        .PERM                  => return LinuxGenericError.OperationNotPermittedError,
        .NOENT                 => return LinuxGenericError.NoSuchFileOrDirectoryError,
        .SRCH                  => return LinuxGenericError.NoSuchProcessError,
        .INTR                  => return LinuxGenericError.InterruptedSystemCallError,
        .IO                    => return LinuxGenericError.IOError,
        .NXIO                  => return LinuxGenericError.NoSuchDeviceOrAddressError,
        .@"2BIG"               => return LinuxGenericError.ArgumentListTooLongError,
        .NOEXEC                => return LinuxGenericError.ExecFormatError,
        .BADF                  => return LinuxGenericError.BadFileNumberError,
        .CHILD                 => return LinuxGenericError.NoChildProcessesError,
        .AGAIN                 => return LinuxGenericError.TryAgainError,
        .NOMEM                 => return LinuxGenericError.OutOfMemoryError,
        .ACCES                 => return LinuxGenericError.PermissionDeniedError,
        .FAULT                 => return LinuxGenericError.BadAddressError,
        .NOTBLK                => return LinuxGenericError.BlockDeviceRequiredError,
        .BUSY                  => return LinuxGenericError.DeviceOrResourceBusyError,
        .EXIST                 => return LinuxGenericError.FileExistsError,
        .XDEV                  => return LinuxGenericError.CrossDeviceLinkError,
        .NODEV                 => return LinuxGenericError.NoSuchDeviceError,
        .NOTDIR                => return LinuxGenericError.NotADirectoryError,
        .ISDIR                 => return LinuxGenericError.IsADirectoryError,
        .INVAL                 => return LinuxGenericError.InvalidArgumentError,
        .NFILE                 => return LinuxGenericError.FileTableOverflowError,
        .MFILE                 => return LinuxGenericError.TooManyOpenFilesError,
        .NOTTY                 => return LinuxGenericError.NotATypewriterError,
        .TXTBSY                => return LinuxGenericError.TextFileBusyError,
        .FBIG                  => return LinuxGenericError.FileTooLargeError,
        .NOSPC                 => return LinuxGenericError.NoSpaceLeftOnDeviceError,
        .SPIPE                 => return LinuxGenericError.IllegalSeekError,
        .ROFS                  => return LinuxGenericError.ReadOnlyFilesystemError,
        .MLINK                 => return LinuxGenericError.TooManyLinksError,
        .PIPE                  => return LinuxGenericError.BrokenPipeError,
        .DOM                   => return LinuxGenericError.MathArgumentOutOfDomainOfFunctionError,
        .RANGE                 => return LinuxGenericError.MathResultNotRepresentableError,
        .DEADLK                => return LinuxGenericError.ResourceDeadLockWouldOccurError,
        .NAMETOOLONG           => return LinuxGenericError.FileNameTooLongError,
        .NOLCK                 => return LinuxGenericError.NoRecordLocksAvailableError,
        .NOSYS                 => return LinuxGenericError.FunctionNotImplementedError,
        .NOTEMPTY              => return LinuxGenericError.DirectoryNotEmptyError,
        .LOOP                  => return LinuxGenericError.TooManySymbolicLinksEncounteredError,
        .NOMSG                 => return LinuxGenericError.NoMessageOfDesiredTypeError,
        .IDRM                  => return LinuxGenericError.IdentifierRemovedError,
        .CHRNG                 => return LinuxGenericError.ChannelNumberOutOfRangeError,
        .L2NSYNC               => return LinuxGenericError.Level2NotSynchronizedError,
        .L3HLT                 => return LinuxGenericError.Level3HaltedError,
        .L3RST                 => return LinuxGenericError.Level3ResetError,
        .LNRNG                 => return LinuxGenericError.LinkNumberOutOfRangeError,
        .UNATCH                => return LinuxGenericError.ProtocolDriverNotAttachedError,
        .NOCSI                 => return LinuxGenericError.NoCSIStructureAvailableError,
        .L2HLT                 => return LinuxGenericError.Level2HaltedError,
        .BADE                  => return LinuxGenericError.InvalidExchangeError,
        .BADR                  => return LinuxGenericError.InvalidRequestDescriptorError,
        .XFULL                 => return LinuxGenericError.ExchangeFullError,
        .NOANO                 => return LinuxGenericError.NoAnodeError,
        .BADRQC                => return LinuxGenericError.InvalidRequestCodeError,
        .BADSLT                => return LinuxGenericError.InvalidSlotError,
        .BFONT                 => return LinuxGenericError.BadFontFileFormatError,
        .NOSTR                 => return LinuxGenericError.DeviceNotAStreamError,
        .NODATA                => return LinuxGenericError.NoDataAvailableError,
        .TIME                  => return LinuxGenericError.TimerExpiredError,
        .NOSR                  => return LinuxGenericError.OutOfStreamsResourcesError,
        .NONET                 => return LinuxGenericError.MachineIsNotOnTheNetworkError,
        .NOPKG                 => return LinuxGenericError.PackageNotInstalledError,
        .REMOTE                => return LinuxGenericError.ObjectIsRemoteError,
        .NOLINK                => return LinuxGenericError.LinkHasBeenSeveredError,
        .ADV                   => return LinuxGenericError.AdvertiseErrorError,
        .SRMNT                 => return LinuxGenericError.SrmountErrorError,
        .COMM                  => return LinuxGenericError.CommunicationErrorOnSendError,
        .PROTO                 => return LinuxGenericError.ProtocolErrorError,
        .MULTIHOP              => return LinuxGenericError.MultihopAttemptedError,
        .DOTDOT                => return LinuxGenericError.RFSSpecificErrorError,
        .BADMSG                => return LinuxGenericError.NotADataMessageError,
        .OVERFLOW              => return LinuxGenericError.ValueTooLargeForDefinedDataTypeError,
        .NOTUNIQ               => return LinuxGenericError.NameNotUniqueOnNetworkError,
        .BADFD                 => return LinuxGenericError.FileDescriptorInBadStateError,
        .REMCHG                => return LinuxGenericError.RemoteAddressChangedError,
        .LIBACC                => return LinuxGenericError.CanNotAccessANeededSharedLibraryError,
        .LIBBAD                => return LinuxGenericError.AccessingACorruptedSharedLibraryError,
        .LIBSCN                => return LinuxGenericError.LibSectionInAOoutCorruptedError,
        .LIBMAX                => return LinuxGenericError.AttemptingToLinkInTooManySharedLibrariesError,
        .LIBEXEC               => return LinuxGenericError.CannotExecASharedLibraryDirectlyError,
        .ILSEQ                 => return LinuxGenericError.IllegalByteSequenceError,
        .RESTART               => return LinuxGenericError.InterruptedSystemCallShouldBeRestartedError,
        .STRPIPE               => return LinuxGenericError.StreamsPipeErrorError,
        .USERS                 => return LinuxGenericError.TooManyUsersError,
        .NOTSOCK               => return LinuxGenericError.SocketOperationOnNonSocketError,
        .DESTADDRREQ           => return LinuxGenericError.DestinationAddressRequiredError,
        .MSGSIZE               => return LinuxGenericError.MessageTooLongError,
        .PROTOTYPE             => return LinuxGenericError.ProtocolWrongTypeForSocketError,
        .NOPROTOOPT            => return LinuxGenericError.ProtocolNotAvailableError,
        .PROTONOSUPPORT        => return LinuxGenericError.ProtocolNotSupportedError,
        .SOCKTNOSUPPORT        => return LinuxGenericError.SocketTypeNotSupportedError,
        .OPNOTSUPP             => return LinuxGenericError.OperationNotSupportedOnTransportEndpointError,
        .PFNOSUPPORT           => return LinuxGenericError.ProtocolFamilyNotSupportedError,
        .AFNOSUPPORT           => return LinuxGenericError.AddressFamilyNotSupportedByProtocolError,
        .ADDRINUSE             => return LinuxGenericError.AddressAlreadyInUseError,
        .ADDRNOTAVAIL          => return LinuxGenericError.CannotAssignRequestedAddressError,
        .NETDOWN               => return LinuxGenericError.NetworkIsDownError,
        .NETUNREACH            => return LinuxGenericError.NetworkIsUnreachableError,
        .NETRESET              => return LinuxGenericError.NetworkDroppedConnectionBecauseOfResetError,
        .CONNABORTED           => return LinuxGenericError.SoftwareCausedConnectionAbortError,
        .CONNRESET             => return LinuxGenericError.ConnectionResetByPeerError,
        .NOBUFS                => return LinuxGenericError.NoBufferSpaceAvailableError,
        .ISCONN                => return LinuxGenericError.TransportEndpointIsAlreadyConnectedError,
        .NOTCONN               => return LinuxGenericError.TransportEndpointIsNotConnectedError,
        .SHUTDOWN              => return LinuxGenericError.CannotSendAfterTransportEndpointShutdownError,
        .TOOMANYREFS           => return LinuxGenericError.TooManyReferencesCannotSpliceError,
        .TIMEDOUT              => return LinuxGenericError.ConnectionTimedOutError,
        .CONNREFUSED           => return LinuxGenericError.ConnectionRefusedError,
        .HOSTDOWN              => return LinuxGenericError.HostIsDownError,
        .HOSTUNREACH           => return LinuxGenericError.NoRouteToHostError,
        .ALREADY               => return LinuxGenericError.OperationAlreadyInProgressError,
        .INPROGRESS            => return LinuxGenericError.OperationNowInProgressError,
        .STALE                 => return LinuxGenericError.StaleNFSFileHandleError,
        .UCLEAN                => return LinuxGenericError.StructureNeedsCleaningError,
        .NOTNAM                => return LinuxGenericError.NotAXENIXNamedTypeFileError,
        .NAVAIL                => return LinuxGenericError.NoXENIXSemaphoresAvailableError,
        .ISNAM                 => return LinuxGenericError.IsANamedTypeFileError,
        .REMOTEIO              => return LinuxGenericError.RemoteIOErrorError,
        .DQUOT                 => return LinuxGenericError.QuotaExceededError,
        .NOMEDIUM              => return LinuxGenericError.NoMediumFoundError,
        .MEDIUMTYPE            => return LinuxGenericError.WrongMediumTypeError,
        .CANCELED              => return LinuxGenericError.OperationCanceledError,
        .NOKEY                 => return LinuxGenericError.RequiredKeyNotAvailableError,
        .KEYEXPIRED            => return LinuxGenericError.KeyHasExpiredError,
        .KEYREVOKED            => return LinuxGenericError.KeyHasBeenRevokedError,
        .KEYREJECTED           => return LinuxGenericError.KeyWasRejectedByServiceError,
        .OWNERDEAD             => return LinuxGenericError.OwnerDiedError,
        .NOTRECOVERABLE        => return LinuxGenericError.StateNotRecoverableError,
        .RFKILL                => return LinuxGenericError.OperationNotPossibleDueToRFKillError,
        .HWPOISON              => return LinuxGenericError.MemoryPageHasHardwareErrorError,
        .NSRNODATA             => return LinuxGenericError.DNSServerReturnedAnswerWithNoDataError,
        .NSRFORMERR            => return LinuxGenericError.DNSServerClaimsQueryWasMisformattedError,
        .NSRSERVFAIL           => return LinuxGenericError.DNSServerReturnedGeneralFailureError,
        .NSRNOTFOUND           => return LinuxGenericError.DomainNameNotFoundError,
        .NSRNOTIMP             => return LinuxGenericError.DNSServerDoesNotImplementRequestedOperationError,
        .NSRREFUSED            => return LinuxGenericError.DNSServerRefusedQueryError,
        .NSRBADQUERY           => return LinuxGenericError.MisformattedDNSQueryError,
        .NSRBADNAME            => return LinuxGenericError.MisformattedDomainNameError,
        .NSRBADFAMILY          => return LinuxGenericError.UnsupportedAddressFamilyError,
        .NSRBADRESP            => return LinuxGenericError.MisformattedDNSReplyError,
        .NSRCONNREFUSED        => return LinuxGenericError.CouldNotContactDNSServersError,
        .NSRTIMEOUT            => return LinuxGenericError.TimeoutWhileContactingDNSServersError,
        .NSROF                 => return LinuxGenericError.EndOfFileError,
        .NSRFILE               => return LinuxGenericError.ErrorReadingFileError,
        .NSRNOMEM              => return LinuxGenericError.NameServerOutOfMemoryError,
        .NSRDESTRUCTION        => return LinuxGenericError.ApplicationTerminatedLookupError,
        .NSRQUERYDOMAINTOOLONG => return LinuxGenericError.DomainNameIsTooLongError,
        .NSRCNAMELOOP          => return LinuxGenericError.NameServerCNameLoopError,
        else                   => unreachable,
    }
    // zig fmt: on
}

pub const LinuxGenericError = error{
    OperationNotPermittedError,
    NoSuchFileOrDirectoryError,
    NoSuchProcessError,
    InterruptedSystemCallError,
    IOError,
    NoSuchDeviceOrAddressError,
    ArgumentListTooLongError,
    ExecFormatError,
    BadFileNumberError,
    NoChildProcessesError,
    TryAgainError,
    OutOfMemoryError,
    PermissionDeniedError,
    BadAddressError,
    BlockDeviceRequiredError,
    DeviceOrResourceBusyError,
    FileExistsError,
    CrossDeviceLinkError,
    NoSuchDeviceError,
    NotADirectoryError,
    IsADirectoryError,
    InvalidArgumentError,
    FileTableOverflowError,
    TooManyOpenFilesError,
    NotATypewriterError,
    TextFileBusyError,
    FileTooLargeError,
    NoSpaceLeftOnDeviceError,
    IllegalSeekError,
    ReadOnlyFilesystemError,
    TooManyLinksError,
    BrokenPipeError,
    MathArgumentOutOfDomainOfFunctionError,
    MathResultNotRepresentableError,
    ResourceDeadLockWouldOccurError,
    FileNameTooLongError,
    NoRecordLocksAvailableError,
    FunctionNotImplementedError,
    DirectoryNotEmptyError,
    TooManySymbolicLinksEncounteredError,
    NoMessageOfDesiredTypeError,
    IdentifierRemovedError,
    ChannelNumberOutOfRangeError,
    Level2NotSynchronizedError,
    Level3HaltedError,
    Level3ResetError,
    LinkNumberOutOfRangeError,
    ProtocolDriverNotAttachedError,
    NoCSIStructureAvailableError,
    Level2HaltedError,
    InvalidExchangeError,
    InvalidRequestDescriptorError,
    ExchangeFullError,
    NoAnodeError,
    InvalidRequestCodeError,
    InvalidSlotError,
    BadFontFileFormatError,
    DeviceNotAStreamError,
    NoDataAvailableError,
    TimerExpiredError,
    OutOfStreamsResourcesError,
    MachineIsNotOnTheNetworkError,
    PackageNotInstalledError,
    ObjectIsRemoteError,
    LinkHasBeenSeveredError,
    AdvertiseErrorError,
    SrmountErrorError,
    CommunicationErrorOnSendError,
    ProtocolErrorError,
    MultihopAttemptedError,
    RFSSpecificErrorError,
    NotADataMessageError,
    ValueTooLargeForDefinedDataTypeError,
    NameNotUniqueOnNetworkError,
    FileDescriptorInBadStateError,
    RemoteAddressChangedError,
    CanNotAccessANeededSharedLibraryError,
    AccessingACorruptedSharedLibraryError,
    LibSectionInAOoutCorruptedError,
    AttemptingToLinkInTooManySharedLibrariesError,
    CannotExecASharedLibraryDirectlyError,
    IllegalByteSequenceError,
    InterruptedSystemCallShouldBeRestartedError,
    StreamsPipeErrorError,
    TooManyUsersError,
    SocketOperationOnNonSocketError,
    DestinationAddressRequiredError,
    MessageTooLongError,
    ProtocolWrongTypeForSocketError,
    ProtocolNotAvailableError,
    ProtocolNotSupportedError,
    SocketTypeNotSupportedError,
    OperationNotSupportedOnTransportEndpointError,
    ProtocolFamilyNotSupportedError,
    AddressFamilyNotSupportedByProtocolError,
    AddressAlreadyInUseError,
    CannotAssignRequestedAddressError,
    NetworkIsDownError,
    NetworkIsUnreachableError,
    NetworkDroppedConnectionBecauseOfResetError,
    SoftwareCausedConnectionAbortError,
    ConnectionResetByPeerError,
    NoBufferSpaceAvailableError,
    TransportEndpointIsAlreadyConnectedError,
    TransportEndpointIsNotConnectedError,
    CannotSendAfterTransportEndpointShutdownError,
    TooManyReferencesCannotSpliceError,
    ConnectionTimedOutError,
    ConnectionRefusedError,
    HostIsDownError,
    NoRouteToHostError,
    OperationAlreadyInProgressError,
    OperationNowInProgressError,
    StaleNFSFileHandleError,
    StructureNeedsCleaningError,
    NotAXENIXNamedTypeFileError,
    NoXENIXSemaphoresAvailableError,
    IsANamedTypeFileError,
    RemoteIOErrorError,
    QuotaExceededError,
    NoMediumFoundError,
    WrongMediumTypeError,
    OperationCanceledError,
    RequiredKeyNotAvailableError,
    KeyHasExpiredError,
    KeyHasBeenRevokedError,
    KeyWasRejectedByServiceError,
    OwnerDiedError,
    StateNotRecoverableError,
    OperationNotPossibleDueToRFKillError,
    MemoryPageHasHardwareErrorError,
    DNSServerReturnedAnswerWithNoDataError,
    DNSServerClaimsQueryWasMisformattedError,
    DNSServerReturnedGeneralFailureError,
    DomainNameNotFoundError,
    DNSServerDoesNotImplementRequestedOperationError,
    DNSServerRefusedQueryError,
    MisformattedDNSQueryError,
    MisformattedDomainNameError,
    UnsupportedAddressFamilyError,
    MisformattedDNSReplyError,
    CouldNotContactDNSServersError,
    TimeoutWhileContactingDNSServersError,
    EndOfFileError,
    ErrorReadingFileError,
    NameServerOutOfMemoryError,
    ApplicationTerminatedLookupError,
    DomainNameIsTooLongError,
    NameServerCNameLoopError,
};
